# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for the CLI."""

from faker import Faker

from invenio_communities.cli import (
    create_communities_custom_field,
    custom_field_exists_in_communities,
)
from invenio_communities.communities.records.api import Community
from invenio_communities.fixtures.demo import create_fake_community
from invenio_communities.fixtures.tasks import create_demo_community


def test_fake_demo_community_creation(
    app, db, location, search_clear, community_type_record
):
    """Assert that demo community creation works without failing."""
    faker = Faker()
    create_demo_community(create_fake_community(faker))


def test_create_communities_custom_fields(app, location, db, search_clear, cli_runner):
    """Assert that custom fields mappings are created for communities."""
    result = cli_runner(create_communities_custom_field, "-f", "mycommunityfield")
    assert result.exit_code == 0

    community_mapping_field = list(Community.index.get_mapping().values())[0][
        "mappings"
    ]["properties"]["custom_fields"]
    expected_value = {
        "properties": {
            "mycommunityfield": {
                "type": "text",
                "fields": {"keyword": {"type": "keyword"}},
            }
        }
    }
    assert community_mapping_field == expected_value

    # check for existence
    result = cli_runner(custom_field_exists_in_communities, "-f", "mycommunityfield")
    assert result.exit_code == 0
    assert "Field mycommunityfield exists" in result.output

    result = cli_runner(custom_field_exists_in_communities, "-f", "unknownfield")
    assert result.exit_code == 0
    assert "Field unknownfield does not exist" in result.output



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""UI community schema."""

from functools import partial

from flask import g
from flask_resources import BaseObjectSchema
from invenio_records_resources.services.custom_fields import CustomFieldsSchemaUI
from invenio_vocabularies.contrib.awards.serializer import AwardL10NItemSchema
from invenio_vocabularies.contrib.funders.serializer import FunderL10NItemSchema
from invenio_vocabularies.resources import VocabularyL10Schema
from marshmallow import Schema, fields

from invenio_communities.proxies import current_communities


def _community_permission_check(action, community, identity):
    """Check community permission for identity."""
    return current_communities.service.config.permission_policy_cls(
        action,
        community_id=getattr(community, "id", community["id"]),
        record=community,
    ).allows(identity)


class FundingSchema(Schema):
    """Schema for dumping types in the UI."""

    award = fields.Nested(AwardL10NItemSchema)
    funder = fields.Nested(FunderL10NItemSchema)


class UICommunitySchema(BaseObjectSchema):
    """Schema for dumping extra information of the community for the UI."""

    type = fields.Nested(VocabularyL10Schema, attribute="metadata.type")

    funding = fields.List(
        fields.Nested(FundingSchema()),
        attribute="metadata.funding",
    )

    # Custom fields
    custom_fields = fields.Nested(
        partial(CustomFieldsSchemaUI, fields_var="COMMUNITIES_CUSTOM_FIELDS")
    )

    permissions = fields.Method("get_permissions", dump_only=True)

    def get_permissions(self, obj):
        """Get permission."""
        can_include_directly = _community_permission_check(
            "include_directly", community=obj, identity=g.identity
        )
        can_update = _community_permission_check(
            "update", community=obj, identity=g.identity
        )
        return {"can_include_directly": can_include_directly, "can_update": can_update}


class TypesSchema(Schema):
    """Schema for dumping types in the UI."""

    types = fields.List(fields.Nested(VocabularyL10Schema), attribute="types")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test community member service."""

import pytest
from invenio_access.permissions import system_identity
from invenio_accounts.proxies import current_datastore
from invenio_cache import current_cache
from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_requests.records.api import RequestEvent
from marshmallow import ValidationError

from invenio_communities.members.errors import AlreadyMemberError, InvalidMemberError
from invenio_communities.members.records.api import ArchivedInvitation, Member


#
# Add members
#
@pytest.mark.parametrize(
    "actor,role",
    [
        ("owner", "owner"),
        ("owner", "manager"),
        ("owner", "curator"),
        ("owner", "reader"),
        ("manager", "manager"),
        ("manager", "curator"),
        ("manager", "reader"),
    ],
)
def test_add_allowed(member_service, community, members, group, actor, role, db):
    """Test that the given roles CAN add a group member."""
    current_cache.clear()
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": role,
    }
    member_service.add(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize(
    "actor,role",
    [
        ("manager", "owner"),
        ("curator", "owner"),
        ("curator", "manager"),
        ("curator", "curator"),
        ("curator", "reader"),
        ("reader", "owner"),
        ("reader", "manager"),
        ("reader", "curator"),
        ("reader", "reader"),
    ],
)
def test_add_denied(member_service, community, members, group, actor, role, db):
    """Test that the given roles CANNOT add a group member."""
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": role,
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.add,
        members[actor].identity,
        community._record.id,
        data,
    )


def test_add_duplicate(member_service, community, owner, group, db):
    """Test that we cannot add duplicate members."""
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    member_service.add(owner.identity, community._record.id, data)
    data["role"] = "curator"
    # Duplicate detected (even if role changed)
    assert pytest.raises(
        AlreadyMemberError,
        member_service.add,
        owner.identity,
        community._record.id,
        data,
    )


def test_add_invalid_member_type(member_service, community, owner, new_user, db):
    """Only system identity can add a denied member type."""
    data = {
        # "type: user" is allowed to add only by system
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.add,
        owner.identity,
        community._record.id,
        data,
    )
    member_service.add(system_identity, community._record.id, data)


def test_add_visible_property(member_service, community, owner, new_user, db):
    """Only system identity can set visible property."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
        # visible allowed only to be set by system.
        "visible": True,
    }
    assert pytest.raises(
        ValidationError, member_service.add, owner.identity, community._record.id, data
    )
    member_service.add(system_identity, community._record.id, data)


def test_add_invalid_data(member_service, community, owner, group, db):
    """Test various forms of invalid data."""
    # Invalid group id
    data = {"members": [{"type": "group", "id": "invalid"}], "role": "reader"}
    assert pytest.raises(
        InvalidMemberError,
        member_service.add,
        owner.identity,
        community._record.id,
        data,
    )
    # Missing member id
    data = {"members": [{"type": "group"}], "role": "reader"}
    assert pytest.raises(
        ValidationError, member_service.add, owner.identity, community._record.id, data
    )
    # Missing member type
    data = {"members": [{"id": "groupname"}], "role": "reader"}
    assert pytest.raises(
        ValidationError, member_service.add, owner.identity, community._record.id, data
    )
    # Invalid member type
    data = {"members": [{"type": "invalid", "id": "1"}], "role": "reader"}
    assert pytest.raises(
        ValidationError, member_service.add, owner.identity, community._record.id, data
    )
    # No members
    data = {"members": [], "role": "reader"}
    assert pytest.raises(
        Exception, member_service.add, owner.identity, community._record.id, data
    )
    # Invalid role
    data = {"members": [{"type": "group", "id": group.name}], "role": "inval"}
    assert pytest.raises(
        ValidationError, member_service.add, owner.identity, community._record.id, data
    )
    # Cannot add email
    data = {
        "members": [{"type": "email", "id": "somebody@somewhere.org"}],
        "role": "reader",
    }
    assert pytest.raises(
        ValidationError,
        # using system_identity because owner gets a permission denied before
        # validation kicks in.
        member_service.add,
        system_identity,
        community._record.id,
        data,
    )


#
# Invite members
#
def test_invite(member_service, community, owner, new_user, db):
    """Invite a user."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)
    # ensure that the invited user request has been indexed
    res = member_service.search_invitations(
        owner.identity, community._record.id
    ).to_dict()
    assert res["hits"]["total"] == 1
    # Cannot invite twice.
    pytest.raises(
        AlreadyMemberError,
        member_service.invite,
        owner.identity,
        community._record.id,
        data,
    )
    # Cannot add either if invited
    pytest.raises(
        AlreadyMemberError,
        member_service.add,
        system_identity,
        community._record.id,
        data,
    )


def test_invite_group_denied(member_service, community, owner, group, db):
    """Invite a group."""
    # Groups cannot be invited (groups cannot receive invitation request)
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    # - owner identity
    assert pytest.raises(
        PermissionDeniedError,
        member_service.invite,
        owner.identity,
        community._record.id,
        data,
    )
    # - system identity
    assert pytest.raises(
        InvalidMemberError,
        member_service.invite,
        system_identity,
        community._record.id,
        data,
    )
    # Email invitations not yet implemented
    data["members"][0]["type"] = "email"
    assert pytest.raises(
        ValidationError,
        member_service.invite,
        owner.identity,
        community._record.id,
        data,
    )


def test_invite_already_member(member_service, community, owner, new_user, db):
    """Invite a user."""
    res = member_service.search(owner.identity, community._record.id).to_dict()
    current = res["hits"]["total"]

    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    res = member_service.search(owner.identity, community._record.id).to_dict()
    assert res["hits"]["total"] == current + 1
    # Cannot invite if already a member.
    pytest.raises(
        AlreadyMemberError,
        member_service.invite,
        owner.identity,
        community._record.id,
        data,
    )


def test_invite_with_message(member_service, community, owner, new_user, db):
    """Invite a user."""
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": "reader",
        "message": "Welcome to the club!",
    }
    assert member_service.invite(owner.identity, community._record.id, data)
    # Invalid message
    data["message"] = 1
    pytest.raises(
        ValidationError,
        member_service.invite,
        owner.identity,
        community._record.id,
        data,
    )


def test_invite_view_request(
    events_service, requests_service, invite_user, db, clean_index
):
    """A request should have been created."""
    res = requests_service.search(
        invite_user.identity,
        receiver={"user": invite_user.id},
        type="community-invitation",
    ).to_dict()
    assert res["hits"]["total"] == 1

    # check request comment since invite user has a message
    RequestEvent.index.refresh()
    res = events_service.search(
        invite_user.identity,
        request_id=res["hits"]["hits"][0]["id"],
    ).to_dict()
    hits = res["hits"]
    assert hits["total"] == 2  # role + invitation
    assert hits["hits"][0]["payload"]["content"] == 'You will join as "Reader".'
    assert hits["hits"][1]["payload"]["content"] == "Welcome to the club!"


#
# Search and read members
#
def test_search_members(
    member_service, community, owner, public_reader, anon_identity, clean_index
):
    """Members can see all members, anyone can only see public."""
    # Members can see all other members.
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 2
    # search on the affiliation (tests query expansion)
    res = member_service.search(
        owner.identity, community._record.id, q=f"affiliation:CERN"
    )
    assert res.to_dict()["hits"]["total"] == 1
    res = member_service.search(owner.identity, community._record.id, q=f"name:New")
    assert res.to_dict()["hits"]["total"] == 1
    res = member_service.search(
        owner.identity, community._record.id, q=f"email:newuser@newuser.org"
    )
    assert res.to_dict()["hits"]["total"] == 1


def test_search_public_members(
    member_service, community, owner, public_reader, anon_identity, clean_index
):
    """Members can see all members, anyone can only see public."""
    # Anyone else can only see public members.
    res = member_service.search_public(anon_identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 1
    # search on the affiliation (tests query expansion)
    res = member_service.search_public(
        owner.identity, community._record.id, q=f"affiliation:CERN"
    )
    assert res.to_dict()["hits"]["total"] == 1
    res = member_service.search_public(
        owner.identity, community._record.id, q=f"name:New"
    )
    assert res.to_dict()["hits"]["total"] == 1
    # search on private fields should not get hits
    res = member_service.search_public(
        owner.identity, community._record.id, q=f"newuser@newuser.org"
    )
    assert res.to_dict()["hits"]["total"] == 0
    # should get hits if using private search
    res = member_service.search(
        owner.identity, community._record.id, q=f"newuser@newuser.org"
    )
    assert res.to_dict()["hits"]["total"] == 1


def test_search_members_restricted(
    member_service, restricted_community, owner, anon_identity, clean_index
):
    """Restricted communities can only be searched by members."""
    c = restricted_community

    # Members can see all other members.
    res = member_service.search(owner.identity, c._record.id)
    assert res.to_dict()["hits"]["total"] == 1

    # Anyone get permission denied.
    pytest.raises(
        PermissionDeniedError, member_service.search_public, anon_identity, c._record.id
    )


def test_search_members_restricted_as_group(
    member_service, restricted_community, admin, owner, db, clean_index
):
    """Test if added group members can perform search."""
    data = {
        "members": [{"type": "group", "id": "admin-access"}],
        "role": "owner",
    }

    pytest.raises(
        PermissionDeniedError,
        member_service.search,
        admin.identity,
        restricted_community._record.id,
    )

    # adding group so its members are able to perform actions
    member_service.add(owner.identity, restricted_community._record.id, data)
    admin.refresh()
    res = member_service.search(admin.identity, restricted_community._record.id)
    assert res.total == 2


def test_read_memberships(member_service, community, any_user, db, clean_index):
    # empty at first
    assert member_service.read_memberships(any_user.identity) == {"memberships": []}
    # add membership
    data = {
        "members": [{"type": "user", "id": str(any_user.id)}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    any_user.refresh()
    assert member_service.read_memberships(any_user.identity) == {
        "memberships": [(str(community._record.id), "reader")]
    }


#
# Search invitations
#
def test_search_invitations(
    member_service,
    requests_service,
    community,
    owner,
    invite_user,
    invite_request_id,
    db,
    clean_index,
):
    """Search invitations should include archived invitations."""
    # Decline the invitation and reinvite
    requests_service.execute_action(
        invite_user.identity, invite_request_id, "decline"
    ).to_dict()
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)

    # See invitations in list
    res = member_service.search_invitations(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 2


#
# Accept/decline/cancel invite
#
def test_invite_accept_flow(
    member_service,
    requests_service,
    community,
    owner,
    invite_user,
    invite_request_id,
    db,
    clean_index,
):
    """Invite a user."""
    # Accept request
    request = requests_service.execute_action(
        invite_user.identity, invite_request_id, "accept"
    ).to_dict()

    # See new member in list
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 2

    # See invitation in archived list
    ArchivedInvitation.index.refresh()
    res = member_service.search_invitations(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 1


def test_invite_decline_flow(
    member_service,
    requests_service,
    community,
    owner,
    invite_user,
    invite_request_id,
    db,
    clean_index,
):
    """Invite a user."""
    # Decline request
    request = requests_service.execute_action(
        invite_user.identity, invite_request_id, "decline"
    ).to_dict()

    # Only owner in list
    Member.index.refresh()
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 1

    # It's possible to reinvite
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    member_service.invite(owner.identity, community._record.id, data)


def test_invite_cancel_flow(
    member_service,
    requests_service,
    community,
    owner,
    invite_request_id,
    db,
    clean_index,
):
    """Invite a user."""
    # Cancel request
    request = requests_service.execute_action(
        owner.identity, invite_request_id, "cancel"
    ).to_dict()

    # Only owner in list
    Member.index.refresh()
    res = member_service.search(owner.identity, community._record.id)
    assert res.to_dict()["hits"]["total"] == 1


def test_invite_actions_permissions(
    db,
    requests_service,
    owner,
    any_user,
    members,
    invite_user,
    invite_request_id,
):
    """Actions should be protected."""
    manager = members["manager"]
    curator = members["curator"]
    reader = members["reader"]

    # Only invited user can accept/decline
    for user in [owner, any_user, manager, curator, reader]:
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity,
            invite_request_id,
            "accept",
        )
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity,
            invite_request_id,
            "decline",
        )

    # Only community owners/managers can cancel
    for user in [any_user, curator, reader, invite_user]:
        pytest.raises(
            PermissionDeniedError,
            requests_service.execute_action,
            user.identity,
            invite_request_id,
            "cancel",
        )


#
# Leave community
#
@pytest.mark.parametrize("role", ["manager", "curator", "reader"])
def test_leave_allowed(member_service, community, members, role):
    """Managers, curators and readers can leave."""
    user = members[role]
    data = {"members": [{"type": "user", "id": str(user.id)}]}
    assert member_service.delete(user.identity, community._record.id, data)


def test_leave_single_owner_denied(member_service, community, owner):
    """A single owner cannot leave"""
    data = {"members": [{"type": "user", "id": str(owner.id)}]}
    pytest.raises(
        ValidationError,
        member_service.delete,
        owner.identity,
        community._record.id,
        data,
    )


def test_leave_denied(db, member_service, community, any_user, invite_user):
    """No permission for others"""
    # some test is not undoing membership correctly
    # FIXME when https://github.com/inveniosoftware/pytest-invenio/issues/30
    current_cache.clear()
    any_user.refresh()
    data = {"members": [{"type": "user", "id": str(any_user.id)}]}
    pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        any_user.identity,
        community._record.id,
        data,
    )
    data = {"members": [{"type": "user", "id": str(invite_user.id)}]}
    pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        invite_user.identity,
        community._record.id,
        data,
    )


def test_leave_owner_allowed(member_service, community, owner, group, db):
    """If multiple owners exists, an owner can leave"""
    # Add an owner
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "owner",
    }
    member_service.add(owner.identity, community._record.id, data)

    # Leave
    data = {"members": [{"type": "user", "id": str(owner.id)}]}
    assert member_service.delete(owner.identity, community._record.id, data)


#
# Delete members
#
# TODO: there should be one user owner (group owner not enough).
@pytest.mark.parametrize(
    "actor,role",
    [
        ("manager", "owner"),
        ("curator", "owner"),
        ("curator", "manager"),
        ("curator", "curator"),
        ("curator", "reader"),
        ("reader", "owner"),
        ("reader", "manager"),
        ("reader", "curator"),
        ("reader", "reader"),
    ],
)
def test_delete_denied(member_service, community, members, actor, new_user, role, db):
    """Test that the given roles CANNOT be removed by the actor."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the user
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
    }
    assert pytest.raises(
        PermissionDeniedError,
        member_service.delete,
        members[actor].identity,
        community._record.id,
        data,
    )


@pytest.mark.parametrize(
    "actor,role",
    [
        ("owner", "owner"),
        ("owner", "manager"),
        ("owner", "curator"),
        ("owner", "reader"),
        ("manager", "manager"),
        ("manager", "curator"),
        ("manager", "reader"),
    ],
)
def test_delete_allowed(member_service, community, members, actor, role, new_user, db):
    """Test that the given roles CAN be removed by the actor."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the member again as the actor
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
    }
    member_service.delete(members[actor].identity, community._record.id, data)


def test_delete_member_type_group(member_service, community, owner, group, db):
    """Groups can be removed."""
    # Add a new user with role
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    # Delete the member again
    data = {
        "members": [{"type": "group", "id": group.name}],
    }
    member_service.delete(owner.identity, community._record.id, data)


def test_delete_invalid_member(member_service, community, owner):
    """Invalid members and member types raises an error."""
    data = {"members": [{"type": "group", "id": "invalid"}]}
    pytest.raises(
        InvalidMemberError,
        member_service.delete,
        owner.identity,
        community._record.id,
        data,
    )
    data = {"members": [{"type": "invalid", "id": "1"}]}
    pytest.raises(
        ValidationError,
        member_service.delete,
        owner.identity,
        community._record.id,
        data,
    )


#
# Update self
#
def test_selfupdate_denied(db, member_service, community, any_user, new_user):
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": False,
    }
    for u in [any_user, new_user]:
        pytest.raises(
            PermissionDeniedError,
            member_service.update,
            u.identity,
            community._record.id,
            data,
        )


@pytest.mark.parametrize(
    "actor",
    [
        "owner",
        "manager",
        "curator",
        "reader",
    ],
)
def test_selfupdate_role_denied(member_service, community, members, actor):
    """Nobody can change their own role."""
    user = members[actor]
    data = {
        "members": [{"type": "user", "id": str(user.id)}],
        "role": actor,
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        user.identity,
        community._record.id,
        data,
    )


@pytest.mark.parametrize(
    "actor",
    [
        "owner",
        "manager",
        "curator",
        "reader",
    ],
)
def test_selfupdate_allow_visibility(member_service, community, members, actor, db):
    """All + system identity can change their own visibility to true."""
    user = members[actor]
    data = {
        "members": [{"type": "user", "id": str(user.id)}],
        "visible": True,
    }
    member_service.update(user.identity, community._record.id, data)
    member_service.update(system_identity, community._record.id, data)


#
# Update members
#
@pytest.mark.parametrize(
    "actor,role",
    [
        ("owner", "owner"),
        ("owner", "manager"),
        ("owner", "curator"),
        ("owner", "reader"),
        ("manager", "manager"),
        ("manager", "curator"),
        ("manager", "reader"),
    ],
)
def test_update_public_visibility_denied(
    member_service, community, members, actor, role, new_user, db
):
    """Only members themselves can set public visibility."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": True,
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        members[actor].identity,
        community._record.id,
        data,
    )


def test_update_public_visibility_of_group_allowed(
    member_service, community, owner, group, db
):
    """Group visibility can always be set to true."""
    # Add a new user with role
    data = {
        "members": [{"type": "group", "id": str(group.name)}],
        "role": "reader",
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member
    data = {
        "members": [{"type": "group", "id": str(group.name)}],
        "visible": True,
    }
    member_service.update(owner.identity, community._record.id, data)


@pytest.mark.parametrize(
    "actor,role",
    [
        ("owner", "owner"),
        ("owner", "manager"),
        ("owner", "curator"),
        ("owner", "reader"),
        ("manager", "manager"),
        ("manager", "curator"),
        ("manager", "reader"),
    ],
)
def test_update_hidden_visibility_allowed(
    member_service, community, members, actor, role, new_user, db
):
    """Only members themselves can set public visibility."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": role,
        "visible": True,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with hidden visibility.
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "visible": False,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize(
    "actor,initial_role,new_role",
    [
        ("owner", "owner", "reader"),
        ("owner", "manager", "reader"),
        ("owner", "curator", "manager"),
        ("owner", "reader", "curator"),
        ("manager", "manager", "curator"),
        ("manager", "curator", "manager"),
        ("manager", "reader", "curator"),
    ],
)
def test_update_role_allowed(
    member_service, community, members, actor, initial_role, new_role, new_user, db
):
    """Owners can change role of all, managers all but owners."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with new role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": new_role,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize(
    "actor,initial_role,new_role",
    [
        ("manager", "owner", "reader"),
        ("curator", "owner", "reader"),
        ("curator", "manager", "curator"),
        ("curator", "curator", "manager"),
        ("curator", "reader", "owner"),
        ("reader", "owner", "owner"),
        ("reader", "manager", "manager"),
        ("reader", "curator", "curator"),
        ("reader", "reader", "manager"),
    ],
)
def test_update_role_denied(
    member_service, community, members, actor, initial_role, new_role, new_user, db
):
    """Managers cannot change owner roles, curators/readers cannot change."""
    # Add a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.add(system_identity, community._record.id, data)
    # Update the member with new role
    data["role"] = new_role
    pytest.raises(
        PermissionDeniedError,
        member_service.update,
        members[actor].identity,
        community._record.id,
        data,
    )


@pytest.mark.parametrize(
    "actor,initial_role,new_role",
    [
        ("owner", "owner", "reader"),
        ("owner", "manager", "reader"),
        ("owner", "curator", "manager"),
        ("owner", "reader", "curator"),
        ("manager", "manager", "curator"),
        ("manager", "curator", "manager"),
        ("manager", "reader", "curator"),
    ],
)
def test_update_invitation_role_allowed(
    member_service, community, members, actor, initial_role, new_role, new_user, db
):
    """Owners can change role of all, managers all but owners."""
    # Invite a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.invite(members[actor].identity, community._record.id, data)

    # Update the invitation with new role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": new_role,
    }
    member_service.update(members[actor].identity, community._record.id, data)


@pytest.mark.parametrize(
    "actor,initial_role,new_role",
    [
        ("manager", "owner", "reader"),
        ("curator", "owner", "reader"),
        ("curator", "manager", "curator"),
        ("curator", "curator", "manager"),
        ("curator", "reader", "owner"),
        ("reader", "owner", "owner"),
        ("reader", "manager", "manager"),
        ("reader", "curator", "curator"),
        ("reader", "reader", "manager"),
    ],
)
def test_update_invitation_role_denied(
    member_service, community, members, actor, initial_role, new_role, new_user, db
):
    """Managers cannot change owner roles, curators/readers cannot change."""
    # Invite a new user with role
    data = {
        "members": [{"type": "user", "id": str(new_user.id)}],
        "role": initial_role,
    }
    member_service.invite(members["owner"].identity, community._record.id, data)

    # Update the invitation with new role
    data["role"] = new_role
    pytest.raises(
        PermissionDeniedError,
        member_service.update,
        members[actor].identity,
        community._record.id,
        data,
    )


@pytest.mark.parametrize("action", ["decline", "cancel", "expire"])
def test_update_declined_invitation(
    member_service,
    requests_service,
    community,
    owner,
    invite_user,
    invite_request_id,
    db,
    action,
):
    """A declined/cancelled invitation cannot be updated."""
    requests_service.execute_action(
        system_identity, invite_request_id, action
    ).to_dict()

    # Update the invitation with new role
    data = {
        "members": [{"type": "user", "id": str(invite_user.id)}],
        "role": "reader",
    }
    assert pytest.raises(
        InvalidMemberError,
        member_service.update,
        owner.identity,
        community._record.id,
        data,
    )


def test_update_role_must_have_owner(member_service, community, owner, group, db):
    """There must always be at least one owner."""
    # Add an owner group
    data = {
        "members": [{"type": "group", "id": group.name}],
        "role": "owner",
    }
    member_service.add(system_identity, community._record.id, data)

    # Update the both owners with manager role using system_identity
    data = {
        "members": [
            {"type": "group", "id": group.name},
            {"type": "user", "id": str(owner.id)},
        ],
        "role": "manager",
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )


def test_update_invalid_data(member_service, community, group):
    # No role or visibility
    data = {
        "members": [{"type": "group", "id": group.name}],
    }
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )
    # Invalid type
    data = {"members": [{"type": "email", "id": group.name}], "role": "owner"}
    pytest.raises(
        InvalidMemberError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )
    data = {"members": [{"type": "invalid", "id": group.name}], "role": "owner"}
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )
    # No members
    data = {"members": [], "role": "owner"}
    pytest.raises(
        ValidationError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )
    # Missing member
    data = {"members": [{"type": "group", "id": "invalid"}], "role": "owner"}
    pytest.raises(
        InvalidMemberError,
        member_service.update,
        system_identity,
        community._record.id,
        data,
    )


#
# Change notifications
#
def test_relation_update_propagation(
    app, db, clean_index, public_reader, community, member_service
):
    comm_uuid = community._record.id
    # there is no .read() implementation
    comm_members = member_service.search_public(system_identity, comm_uuid)
    assert comm_members.total == 1

    member = list(comm_members.hits)[0]
    assert member.get("member").get("name") == "New User"

    # update user
    user_id = member["member"]["id"]
    user = current_datastore.get_user(user_id)
    user.user_profile = {"full_name": "Update test", "affiliations": "CERN"}
    current_datastore.commit()

    # check community has been updated
    assert member_service.indexer.process_bulk_queue() == (1, 0)
    Member.index.refresh()
    comm_members = member_service.search_public(system_identity, comm_uuid)
    assert comm_members.total == 1

    member = list(comm_members.hits)[0]
    assert member.get("member").get("name") == "Update test"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
# Copyright (C) 2021 Graz University of Technology.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2022 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community permissions."""

import operator
from collections import namedtuple
from functools import partial, reduce
from itertools import chain

from flask import current_app
from flask_principal import UserNeed
from invenio_access.permissions import any_user, system_process
from invenio_records_permissions.generators import Generator
from invenio_search.engine import dsl

from .proxies import current_roles

_Need = namedtuple("Need", ["method", "value", "role"])

CommunityRoleNeed = partial(_Need, "community")
"""Defines a need for a community role.

You can create a community role need like below:

.. code-block:: python

    need = CommunityRoleNeed(community.id, 'manager')
    # Community role needs can be identified by their method attribute
    assert need.method == 'community
    # The community identifier can be accessed via the value attribute
    assert need.value == community.id
    # The role held within the community can be accessed via the role attribute
    assert need.role == 'manager
"""


# TODO: Move class to Invenio-Records-Permissions and make more reusable
class IfRestrictedBase(Generator):
    """IfRestricted generator.

    IfRestricted(
        'record', 'restricted', 'public'
        then_=[RecordPermissionLevel('view')],
        else_=[ActionNeed(superuser-access)],
    )
    """

    def __init__(self, field_getter, field_name, then_value, else_value, then_, else_):
        """Constructor."""
        self.field_getter = field_getter
        self.field_name = field_name
        self.then_value = then_value
        self.else_value = else_value
        self.then_ = then_
        self.else_ = else_

    def generators(self, record):
        """Get the "then" or "else" generators."""
        if record is None:
            # TODO - when permissions on links are checked, the record is not
            # passes properly, causing below ``record.access`` to fail.
            return self.else_
        # TODO: Next statement should be made more reusable
        is_then_value = self.field_getter(record) or self.then_value
        return self.then_ if is_then_value == self.then_value else self.else_

    def needs(self, record=None, **kwargs):
        """Set of Needs granting permission."""
        needs = [g.needs(record=record, **kwargs) for g in self.generators(record)]
        return set(chain.from_iterable(needs))

    def excludes(self, record=None, **kwargs):
        """Set of Needs denying permission."""
        needs = [g.excludes(record=record, **kwargs) for g in self.generators(record)]
        return set(chain.from_iterable(needs))

    def make_query(self, generators, **kwargs):
        """Make a query for one set of generators."""
        queries = [g.query_filter(**kwargs) for g in generators]
        queries = [q for q in queries if q]
        return reduce(operator.or_, queries) if queries else None

    def query_filter(self, **kwargs):
        """Filters for current identity."""
        # TODO: Next two statement should be made more reusable
        q_then = dsl.Q("match", **{self.field_name: self.then_value})
        q_else = dsl.Q("match", **{self.field_name: self.else_value})
        then_query = self.make_query(self.then_, **kwargs)
        else_query = self.make_query(self.else_, **kwargs)

        if then_query and else_query:
            return (q_then & then_query) | (q_else & else_query)
        elif then_query:
            return (q_then & then_query) | q_else
        elif else_query:
            return q_else & else_query
        else:
            return q_else


# TODO: Remove class once IfRestrictedBase has been moved.
class IfRestricted(IfRestrictedBase):
    """IfRestricted."""

    def __init__(self, field, then_, else_):
        """Initialize."""
        super().__init__(
            lambda r: getattr(r.access, field, None),
            f"access.{field}",
            "restricted",
            "public",
            then_,
            else_,
        )


class IfPolicyClosed(IfRestrictedBase):
    """If policy is closed."""

    def __init__(self, field, then_, else_):
        """Initialize."""
        super().__init__(
            lambda r: getattr(r.access, field, None)
            if hasattr(r, "access")
            else r.get("access", {}).get(
                field
            ),  # needed for running permission check at serialization time and avoid db query
            f"access.{field}",
            "closed",
            "open",
            then_,
            else_,
        )


#
# Community membership generators
#
class CommunityRoles(Generator):
    """Base class for community roles generators."""

    def roles(self, **kwargs):
        """R."""
        raise NotImplementedError

    def communities(self, identity):
        """Communities."""
        raise NotImplementedError

    def needs(self, record=None, community_id=None, **kwargs):
        """Enabling Needs."""
        if community_id is None:
            community_id = record.id

        assert community_id, "No community id provided."
        community_id = str(community_id)

        roles = self.roles(**kwargs)
        if roles:
            needs = [CommunityRoleNeed(community_id, role) for role in roles]
            return needs
        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        # Gives access to all community members.
        return dsl.Q("terms", **{"_id": self.communities(identity)})


class CommunityMembers(CommunityRoles):
    """Roles representing all members of a community."""

    def roles(self, **kwargs):
        """Roles."""
        return [r.name for r in current_roles]

    def communities(self, identity):
        """Communities."""
        return [n.value for n in identity.provides if n.method == "community"]


class CommunityCurators(CommunityRoles):
    """Roles representing all curators of a community."""

    def roles(self, **kwargs):
        """Roles."""
        return [r.name for r in current_roles.can("curate")]


class CommunityManagers(CommunityRoles):
    """Roles representing all managers of a community."""

    def roles(self, **kwargs):
        """Roles."""
        return [r.name for r in current_roles.can("manage")]


class CommunityManagersForRole(CommunityRoles):
    """Roles representing all managers of a community for a role update."""

    def roles(self, role=None, member=None, **kwargs):
        """Roles."""
        allowed_roles = []
        if role is not None and member is not None:
            # Update from an old role to a new role. The most restrictive set
            # applies.
            new_allowed_roles = set(current_roles.manager_roles(role))
            old_allowed_roles = set(current_roles.manager_roles(member.role))
            allowed_roles = new_allowed_roles & old_allowed_roles
        elif role is not None:
            # Adding/inviting a new role
            allowed_roles = current_roles.manager_roles(role)
        elif member is not None:
            # Deleting or updating a member with a given role (without changing
            # role)
            allowed_roles = current_roles.manager_roles(member.role)
        else:
            raise NotImplementedError("You must provide a role and/or a member.")

        return [r.name for r in allowed_roles]


class CommunityOwners(CommunityRoles):
    """Roles representing the owners of a community."""

    def roles(self, **kwargs):
        """Roles."""
        return [current_roles.owner_role.name]

    def communities(self, identity):
        """Communities."""
        return [
            n.value
            for n in identity.provides
            if n.method == "community" and n.role == current_roles.owner_role.name
        ]


class CommunitySelfMember(Generator):
    """Allows a member themselves (only for users)."""

    def needs(self, member=None, **kwargs):
        """Enabling needs."""
        if member is not None and member.user_id is not None:
            return [UserNeed(member.user_id)]
        return []

    def query_filter(self, identity=None, **kwargs):
        """Not implemented."""
        raise NotImplementedError("Search permission not supported.")


#
# Business-level rules.
#
class AllowedMemberTypes(Generator):
    """Generator to restrict to set of allowed member types.

    We are restricting member types that can be added/invited via permissions
    so that it's easily overridable for an instance to change behavior.

    A system process is allowed to do anything.
    """

    def __init__(self, *allowed_member_types):
        """Allowed member types."""
        self.allowed_member_types = allowed_member_types

    def needs(self, **kwargs):
        """Enabling needs."""
        return [system_process]

    def excludes(self, member_types=None, **kwargs):
        """Preventing needs."""
        if member_types:
            for m in member_types:
                if m not in self.allowed_member_types:
                    return [any_user]
        return []


class GroupsEnabled(Generator):
    """Generator to restrict if the groups are not enabled.

    If the groups are not enabled, exclude any user for adding members of the
    param member type.

    A system process is allowed to do anything.
    """

    def __init__(self, *need_groups_enabled_types):
        """Types that need the groups enabled."""
        self.need_groups_enabled_types = need_groups_enabled_types

    def excludes(self, member_types=None, **kwargs):
        """Preventing needs."""
        if member_types:
            for m in member_types:
                if (
                    m in self.need_groups_enabled_types
                    and not current_app.config["COMMUNITIES_GROUPS_ENABLED"]
                ):
                    return [any_user]
        return []



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
# Copyright (C) 2022 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import pytest
from flask_security import login_user
from invenio_accounts.testutils import login_user_via_session
from invenio_records_resources.proxies import current_service_registry
from invenio_vocabularies.contrib.affiliations.api import Affiliation


@pytest.fixture()
def affiliation(app, db, superuser_identity):
    """Affiliation vocabulary record."""
    aff = current_service_registry.get("affiliations").create(
        superuser_identity,
        {
            "name": "CERN",
            "id": "cern",
            "acronym": "CERN",
            "identifiers": [
                {
                    "scheme": "ror",
                    "identifier": "01ggx4157",
                },
                {
                    "scheme": "isni",
                    "identifier": "000000012156142X",
                },
            ],
        },
    )

    Affiliation.index.refresh()

    return aff


@pytest.fixture()
def client_with_login(client, users):
    """Log in a user to the client."""
    user = users[0]
    login_user(user, remember=True)
    login_user_via_session(client, email=user.email)
    return client
